use std::error::Error;
use std::fmt;
use std::fs;
use std::path::Path;

mod reader;

#[cfg(windows)]
const LINE_ENDING: &str = "\r\n";
#[cfg(not(windows))]
const LINE_ENDING: &str = "\n";

#[derive(PartialEq, Copy, Clone)]
enum CommentStyle {
    Semicolon,
    Hash,
}

type CommentLine = (CommentStyle, String);

struct CommentBlock {
    style: CommentStyle,
    lines: Vec<String>,
}

const COMMENT_STYLE_TO_PREFIX_MAP: [(CommentStyle, &str); 2] =
    [(CommentStyle::Semicolon, ";"), (CommentStyle::Hash, "#")];

enum Value {
    Unknown(String),
    UnknownArray(Vec<String>),
}

struct Variable {
    name: String,
    value: Value,
    comments: Option<Vec<CommentBlock>>,
    inline_comment: Option<CommentLine>,
}

enum SectionTitle {
    Class { package: String, class: String },
    PerObject { object_name: String, class: String },
}

struct Section {
    title: SectionTitle,
    variables: Vec<Variable>,
    comments: Option<Vec<CommentBlock>>,
    inline_comment: Option<CommentLine>,
}

pub struct ConfigFile {
    name: String,
    sections: Vec<Section>,
}

impl fmt::Display for CommentBlock {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let prefix = match &self.style {
            CommentStyle::Semicolon => ";",
            CommentStyle::Hash => "#",
        };
        if !self.lines.is_empty() {
            write!(f, "{}{}", prefix, self.lines[0])?;
        }
        for line in self.lines.iter().skip(1) {
            write!(f, "{}{}{}", LINE_ENDING, prefix, line)?;
        }
        Ok(())
    }
}

impl fmt::Display for Value {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Value::Unknown(raw_value) => write!(f, "{}", raw_value)?,
            Value::UnknownArray(raw_values) => {
                write!(f, "[")?;
                if !raw_values.is_empty() {
                    write!(f, "{}", raw_values[0])?;
                }
                for raw_value in raw_values.iter().skip(1) {
                    write!(f, ", {}", raw_value)?;
                }
                write!(f, "]")?;
            }
        };
        Ok(())
    }
}

impl fmt::Display for Variable {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(comment_blocks) = &self.comments {
            format_comment_blocks(f, &comment_blocks)?;
        }
        match &self.value {
            Value::Unknown(raw_value) => write!(f, "{}={}", self.name, raw_value)?,
            Value::UnknownArray(raw_values) => {
                if !raw_values.is_empty() {
                    write!(f, "{}={}", self.name, raw_values[0])?;
                }
                for raw_value in raw_values.iter().skip(1) {
                    write!(f, "{}{}={}", LINE_ENDING, self.name, raw_value)?;
                }
            }
        }
        if let Some(comment) = &self.inline_comment {
            format_comment_line(f, &comment)?;
        }
        Ok(())
    }
}

impl fmt::Display for Section {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        if let Some(comment_blocks) = &self.comments {
            format_comment_blocks(f, &comment_blocks)?;
        }
        match &self.title {
            SectionTitle::Class { package, class } => {
                write!(f, "[{}.{}]", package, class)?;
            }
            SectionTitle::PerObject { object_name, class } => {
                write!(f, "[{} {}]", object_name, class)?;
            }
        }
        if let Some(comment) = &self.inline_comment {
            format_comment_line(f, &comment)?;
        }
        writeln!(f, "")?;
        for variable in self.variables.iter() {
            writeln!(f, "{}", variable)?;
        }
        Ok(())
    }
}

impl fmt::Display for ConfigFile {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        writeln!(f, r#"Config file: "{}""#, self.name)?;
        if self.sections.len() > 0 {
            write!(f, "{}", self.sections[0])?;
        }
        for section in self.sections.iter().skip(1) {
            write!(f, "{}{}", LINE_ENDING, section)?;
        }
        Ok(())
    }
}

fn format_comment_line(f: &mut fmt::Formatter<'_>, (style, content): &CommentLine) -> fmt::Result {
    match style {
        CommentStyle::Semicolon => write!(f, " ;")?,
        CommentStyle::Hash => write!(f, " #")?,
    };
    write!(f, "{}", content)?;
    Ok(())
}

fn format_comment_blocks(f: &mut fmt::Formatter<'_>, blocks: &Vec<CommentBlock>) -> fmt::Result {
    for block in blocks.iter() {
        if block.style == CommentStyle::Hash {
            write!(f, "{}", LINE_ENDING)?;
        }
        write!(f, "{}{}", block, LINE_ENDING)?;
    }
    Ok(())
}

pub fn load_file(file_path: &Path) -> Result<ConfigFile, Box<dyn Error>> {
    let config_name = file_path
        .file_stem()
        .and_then(|x| x.to_str())
        .unwrap_or_default()
        .to_string();
    Ok(fs::read_to_string(file_path)?
        .lines()
        .fold(reader::Reader::new(config_name), |reader, line| {
            reader.add_line(line.to_string())
        })
        .return_config())
}
